// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of editex;

class CursorList {
  final List<int> _indices;
  CursorList([List<int> indices]) : _indices = indices ?? [0];

  /// Used only for rendering selection color.
  var _atCursor = 0;
  static var selectionColor = '#0d47a1';

  CursorList clone() => new CursorList(new List<int>.from(_indices));

  /// Get number of cursors.
  int get count => _indices.length;

  /// Get color stop in selection.
  String getColorStop() {
    _atCursor++;
    if (_atCursor % 2 == 1 && _indices.length - _atCursor > 0) {
      return '\\color{$selectionColor}{';
    } else if (_atCursor % 2 == 0 && _indices.length - _atCursor >= 0) {
      return '}';
    } else {
      return '';
    }
  }

  /// Set to one cursor with the given index.
  void set(int index) {
    _indices
      ..clear()
      ..add(index);
  }

  /// Load given cursor indices.
  void load(List<int> indices) {
    _indices
      ..clear()
      ..addAll(indices);
  }

  /// Get index of single cursor (ignores all secondary cursors but does not
  /// remove them).
  int get single => _indices.first;

  /// Returns cursor index at [i] in a wrapper.
  W<int> w(int i) {
    return new W<int>(_indices.length > i ? _indices[i] : single);
  }

  void decrement() {
    for (var i = 0; i < _indices.length; i++) {
      _indices[i]--;
    }
  }

  bool currentIsCursor() => _indices.any((idx) => idx == 0);

  /// Move cursor at [i] by [d]. First sets cursor to the index of the first
  /// cursor if it doesn't exist already. If the first cursor is moved all
  /// others are automatically removed.
  int move(int i, int d) {
    if (i == 0) {
      set(single);
    } else if (_indices.length < i + 1) {
      do {
        _indices.add(single);
      } while (_indices.length < i + 1);
    }

    _indices[i] += d;
    final ret = _indices[i];

    // If there are multiple cursors at the same spot, eliminate those.
    final others = new Set<int>();
    _indices.removeWhere((idx) {
      if (others.contains(idx)) {
        return true;
      } else {
        others.add(idx);
        return false;
      }
    });

    return ret;
  }
}
