// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of editex;

/// For comparing fragment data.
class FragmentDataEquality extends DefaultEquality<dynamic> {
  const FragmentDataEquality();

  @override
  bool equals(a, b) {
    if (a is String) {
      return b is String && a == b;
    } else if (a is List) {
      return b is List &&
          const ListEquality(const FragmentDataEquality()).equals(a, b);
    } else {
      return false;
    }
  }
}

/// For [EdiTeX.undoHistory] and [EdiTeX.redoHistory].
class EditHistoryEntry {
  final MultiFragment content;
  final List<int> cursors;
  EditHistoryEntry(this.content, this.cursors);
}

/// EdiTeX editor component.
/// Can initialize itself in an empty <div>
class EdiTeX {
  final EdiTeXInterface interface;
  final DivElement container, katexWrapper;

  var content = new MultiFragment();
  var readonly = false;

  final undoHistory = new List<EditHistoryEntry>();
  final redoHistory = new List<EditHistoryEntry>();
  final selection = new List<Fragment>();
  final cursors = new CursorList();

  var _doUpdate = true;
  var _wrapperWidth;

  final _onLeftLeave = new StreamController<Null>.broadcast();
  final _onRightLeave = new StreamController<Null>.broadcast();
  final _onUpdate = new StreamController<Null>.broadcast();

  EdiTeX(this.container,
      [this.interface = const StatecEdiTeXInterface(
          _defaultCommandResolver, _defaultInstantCommandResolver)])
      : katexWrapper = new DivElement() {
    container.append(katexWrapper);

    // Make focussable via keyboard.
    container.tabIndex = 0;

    // Add keyboard listeners.
    container.onKeyDown.listen(onKeyDown);

    // Finish command on unfocus.
    container.onBlur.listen((_) {
      _finishCurrentKeyword(true);
      _doUpdate = true;

      // If there are multiple cursors, remove them.
      if (cursors.count > 1) {
        cursors.move(0, 0);
      }
    });

    // Update wrapperWidth on.
    _wrapperWidth = katexWrapper.getBoundingClientRect().width;
    window.onResize.listen((_) {
      _wrapperWidth = katexWrapper.getBoundingClientRect().width;
    });

    // Setup rendering loop.
    // Minimal interval is 33ms, this gives a framerate of ~30fps.
    new Timer.periodic(new Duration(milliseconds: 33), (t) {
      if (_doUpdate) {
        _update();
      }
    });
  }

  Future destroy() {
    return Future.wait([_onLeftLeave.close(), _onRightLeave.close()]);
  }

  // Render equation.
  void _update() {
    // Set [doUpdate] to false and notify [_onUpdate].
    _doUpdate = false;
    _onUpdate.add(null);

    // Render content using KaTeX.
    // Note that this code throws an exception if any of the LaTeX is invalid.
    // This should only be the case if a provided template is bad.
    katexWrapper.children.clear();
    katex.render(getLaTeX(), katexWrapper,
        new katex.RenderingOptions(displayMode: true));

    // Format cursor and placeholders.
    Element cursor;
    katexWrapper.getElementsByClassName('mord mathit').forEach((node) {
      final Element element = node;
      if (element.style.color == 'transparent') {
        final id = element.childNodes.first.text;
        element.childNodes.first.remove();
        if (id == 'a') {
          element.append(new SpanElement()..classes.add('editex-cursor'));
          cursor = element;
        } else {
          element.append(new SpanElement()..classes.add('editex-placeholder'));
        }
      }
    });

    // Compute katexDiv offset using cursor position
    final katexDisplay = katexWrapper.children.first;
    final katexLeft = katexDisplay.getBoundingClientRect().left;
    var cursorOffset = cursor.getBoundingClientRect().left;
    cursorOffset -= katexLeft - 1; // -1 because of cursor margin.

    if (cursorOffset > _wrapperWidth) {
      katexDisplay.style.marginLeft = '-${cursorOffset - _wrapperWidth}px';
    } else {
      katexDisplay.style.marginLeft = '0px';
    }
  }

  Stream<Null> get onLeftLeave => _onLeftLeave.stream;
  Stream<Null> get onRightLeave => _onRightLeave.stream;
  Stream<Null> get onUpdate => _onUpdate.stream;

  bool get isEmpty => content.fragments.isEmpty;
  bool get isNotEmpty => content.fragments.isNotEmpty;

  /// Set cursor position.
  void setCursor(int index) {
    cursors.set(index);
    _doUpdate = true;
  }

  /// Add undo checkpoint.
  void _addUndoCheckpoint([EditHistoryEntry entry]) {
    undoHistory.add(entry ?? getState());
    redoHistory.clear();
  }

  /// Get current editor state.
  EditHistoryEntry getState() => new EditHistoryEntry(
      content.clone(), new List<int>.from(cursors._indices));

  /// Undo last change.
  void undo([bool destructive = false]) {
    if (undoHistory.isNotEmpty) {
      if (!destructive) {
        redoHistory.add(getState());
      }

      final replace = undoHistory.removeLast();
      content = replace.content;
      cursors.set(replace.cursors.first);
      _doUpdate = true;
    }
  }

  /// Redo last change.
  void redo() {
    if (redoHistory.isNotEmpty) {
      undoHistory.add(getState());
      final replace = redoHistory.removeLast();
      content = replace.content;
      cursors.load(replace.cursors);
      _doUpdate = true;
    }
  }

  /// Remove selection.
  bool _removeSelection() {
    if (selection.isNotEmpty) {
      final direction = cursors._indices[0] > cursors._indices[1] ? -1 : 0;
      for (var i = 0; i < selection.length; i++) {
        final result = content.remove(cursors.w(0), direction);
        cursors.move(0, direction * result.count);
      }
      selection.clear();
      return true;
    } else {
      return false;
    }
  }

  /// Get selection as JSON data.
  String get selectionJson {
    final tmp = new MultiFragment()..fragments.addAll(selection);
    return JSON.encode(tmp.getData());
  }

  /// Keyboard handler.
  void onKeyDown(KeyboardEvent event) {
    // Save current editor state.
    final oldState = getState();

    // Clear selection if we do no longer have 2 cursors.
    if (cursors.count == 1) {
      selection.clear();
    }

    // Keystrokes
    if (event.ctrlKey) {
      if (event.keyCode == KeyCode.Z) {
        if (event.shiftKey) {
          redo();
        } else {
          undo();
        }

        // To prevend the terminal code to mess up undo/redo, we handle this
        // here.
        _doUpdate = true;
        event.preventDefault();
        return;
      } else if (event.keyCode == KeyCode.C) {
        if (selection.isNotEmpty) {
          window.localStorage['editex-clipboard'] = selectionJson;
        }
      } else if (event.keyCode == KeyCode.V) {
        if (window.localStorage.containsKey('editex-clipboard')) {
          _removeSelection();
          final List data =
              JSON.decode(window.localStorage['editex-clipboard']);
          final tmp = new MultiFragment()..loadData(data, interface);
          for (final frag in tmp.fragments) {
            content.insert(cursors.w(0), frag, interface);
            cursors.move(0, frag.length);
          }
        }
      }
    }

    // Selection
    else if (event.shiftKey &&
        (event.keyCode == KeyCode.LEFT || event.keyCode == KeyCode.RIGHT)) {
      // Clear selection if there are not yet two cursors.
      if (cursors.count == 1) {
        selection.clear();
      }

      final direction = event.keyCode == KeyCode.LEFT ? -1 : 0;
      final frag = content.getFragmentAtCursor(cursors.w(1), direction);
      if (frag != null) {
        final result = cursors.move(1, (direction * 2 + 1) * frag.length);

        // Add fragment to selection.
        if (direction == -1) {
          if (result < cursors.single) {
            selection.insert(0, frag.clone());
          } else {
            selection.removeLast();
          }
        } else {
          if (result > cursors.single) {
            selection.add(frag.clone());
          } else {
            selection.removeAt(0);
          }
        }
      }
    }

    // Normal insert
    else if (new RegExp(r'^[0-9A-Za-z]$').hasMatch(event.key)) {
      // Finish current command if previous character is a number and we now
      // insert a letter.
      if (new RegExp(r'^[A-Za-z]$').hasMatch(event.key)) {
        _finishCurrentKeyword(false, true);
      }

      var char = event.key;
      if (event.shiftKey) {
        char = char.toUpperCase();
      }

      _removeSelection();
      final ret =
          content.insert(cursors.w(0), new CharFragment(char), interface);
      cursors.move(0, ret.cursorOffset);
    }

    // Insert backward slash (command separator).
    else if (event.key == r'\') {
      _finishCurrentKeyword();
      _removeSelection();
      final ret =
          content.insert(cursors.w(0), new CharFragment(r'\'), interface);
      cursors.move(0, ret.cursorOffset);
    }

    // Instant command insert
    else if (interface.hasInstantCommand(event.key)) {
      _finishCurrentKeyword();
      final command = interface.resolveInstantCommand(event.key);
      final fragment = new CommandFragment(command);

      // If the selection is not empty and the command has one argument, put
      // the selection into the command.
      if (selection.isNotEmpty && fragment.args.length == 1) {
        fragment.args[0].fragments
            .addAll(selection.map((frag) => frag.clone()));
      }

      _removeSelection();
      final ret = content.insert(cursors.w(0), fragment, interface);
      cursors.move(0, ret.cursorOffset);
    }

    // Special divide behavior
    else if (event.key == '/') {
      // Only proceed if [_divideCommand] is specified.
      if (interface.hasCommand(_divideCommand)) {
        _finishCurrentKeyword();

        // If there is a selection, use it for the numerator. Else fallback on
        // internal implementation.
        if (selection.isNotEmpty) {
          final command = interface.resolveCommand(_divideCommand);
          final fragment = new CommandFragment(command);
          fragment.args.first.fragments
              .addAll(selection.map((frag) => frag.clone()));
          _removeSelection();
          content.insert(cursors.w(0), fragment, interface);
          cursors.move(0, fragment.length - 1);
        } else {
          final fragment = new CharFragment('/');
          final ret = content.insert(cursors.w(0), fragment, interface);
          cursors.move(0, ret.cursorOffset);
        }
      }
    }

    // Finish a command.
    else if (event.keyCode == KeyCode.SPACE) {
      _finishCurrentKeyword(true);
    }

    // Remove before cursor.
    else if (event.keyCode == KeyCode.BACKSPACE) {
      if (!_removeSelection()) {
        final result = content.remove(cursors.w(0), -1);
        cursors.move(0, -result.count);
      }
    }

    // Remove after cursor.
    else if (event.keyCode == KeyCode.DELETE) {
      if (!_removeSelection()) {
        final result = content.remove(cursors.w(0), 0);
        cursors.move(0, result.removeSelf ? -result.count : 0);
      }
    }

    // Arrow key navigation.
    else if (event.keyCode == KeyCode.LEFT) {
      _finishCurrentKeyword();
      if (cursors.count == 2) {
        cursors.set(min(cursors._indices[0], cursors._indices[1]));
      } else {
        if (cursors.move(0, -1) == -1) {
          cursors.move(0, 1);
          _onLeftLeave.add(null);
        }
      }
    } else if (event.keyCode == KeyCode.RIGHT) {
      _finishCurrentKeyword();
      if (cursors.count == 2) {
        cursors.set(max(cursors._indices[0], cursors._indices[1]));
      } else {
        if (cursors.move(0, 1) == content.length) {
          cursors.move(0, -1);
          _onRightLeave.add(null);
        }
      }
    } else if (event.keyCode == KeyCode.HOME) {
      _finishCurrentKeyword();
      cursors.set(0);
    } else if (event.keyCode == KeyCode.END) {
      _finishCurrentKeyword();
      cursors.set(content.length - 1);
    }

    // Check if the editor state is updated.
    final contentUpdated = !(const FragmentDataEquality()
        .equals(oldState.content.getData(), content.getData()));
    final cursorMoved =
        !(const ListEquality().equals(oldState.cursors, cursors._indices));

    if (contentUpdated) {
      if (readonly) {
        // This it a bit hacky, but very lightweight.
        loadData(oldState.content.getData());
        setCursor(oldState.cursors.first);
      } else {
        _addUndoCheckpoint(oldState);
        _doUpdate = true;
        event.preventDefault();
      }
    } else if (cursorMoved) {
      _doUpdate = true;
      event.preventDefault();
    }
  }

  /// Finished current keyword.
  void _finishCurrentKeyword(
      [bool addUndoCheckpoint = false, bool ifEndsWithDigit = false]) {
    final keywordLength =
        content.finish(cursors.w(0), interface, ifEndsWithDigit);

    // <0 means a keyword was replaced with a command fragment.
    if (keywordLength < 0) {
      // Add one to edit first command parameter (or skip command).
      cursors.move(0, keywordLength + 1);
    }
  }

  /// Get JSON serializable data object.
  List getData() => content.getData();

  /// Load data generated with [getData].
  void loadData(List data) {
    _addUndoCheckpoint();
    content = new MultiFragment()..loadData(data, interface);
    cursors.set(0);
    _doUpdate = true;
  }

  /// Get content as parsable expression.
  String getParsableContent() => content.getParsableContent().trim();

  /// Get LaTeX string with cursor.
  String getLaTeX() {
    return content.render(cursors.clone());
  }
}

class W<T> {
  T v;
  W(this.v);
}
