// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of editex;

class FragmentInsertResult {
  final int cursorOffset;
  final bool didInsert;
  FragmentInsertResult(this.didInsert, [this.cursorOffset = 0]);
}

class FragmentRemoveResult {
  final int count;
  final bool didRemove, removeSelf, dismantleCommand;
  FragmentRemoveResult(this.didRemove,
      {this.count: 0, this.removeSelf: false, this.dismantleCommand: false});
}

/// Expression fragment.
abstract class Fragment {
  Fragment clone();

  String render(CursorList cursors);

  FragmentInsertResult insert(
          W<int> cursor, Fragment fragment, EdiTeXInterface i) =>
      new FragmentInsertResult(false);

  /// [direction]: -1 for backspace and 0 for delete.
  FragmentRemoveResult remove(W<int> cursor, int direction) =>
      new FragmentRemoveResult(false);

  /// If a command is finished this will return the number of cursor positions
  /// that were part of the command.
  int finish(W<int> cursor, EdiTeXInterface i, bool ifEndsWithDigit) => 0;

  /// Get fragment at cursor.
  Fragment getFragmentAtCursor(W<int> cursor, int direction) => null;

  String getParsableContent() => '';

  int get length;
}

/// Fragment containing a single text character.
class CharFragment extends Fragment {
  static const overrideRender = const {r'\': r'\backslash'};

  String char;

  CharFragment(this.char);

  @override
  CharFragment clone() => new CharFragment(char);

  @override
  String render(cursor) {
    if (overrideRender.containsKey(char)) {
      return overrideRender[char];
    }
    return '\\text{$char}';
  }

  @override
  String getParsableContent() => char;

  @override
  int get length => 1;
}

/// Fragment containing a command.
class CommandFragment extends Fragment {
  static final _latexArgRegex =
      new RegExp(r'\$(?:(\d+)|\(([^\d]?)(\d+)([^\d]?)\))');
  static final _parsableArgRegex = new RegExp(r'\$([0-9]+)');

  final EdiTeXCommand data;
  List<MultiFragment> args;

  CommandFragment(this.data, [this.args]) {
    if (args == null) {
      args = new List<MultiFragment>();

      // Create fragment for each argument.
      data.latexTemplate.replaceAllMapped(_latexArgRegex, (_) {
        args.add(new MultiFragment());
        return '';
      });
    }
  }

  @override
  CommandFragment clone() =>
      new CommandFragment(data, args.map((arg) => arg.clone()).toList());

  @override
  FragmentInsertResult insert(cursor, fragment, interface) {
    for (final arg in args) {
      final result = arg.insert(cursor, fragment, interface);
      if (result.didInsert) {
        return result;
      }
      cursor.v--;
    }
    return new FragmentInsertResult(false);
  }

  @override
  FragmentRemoveResult remove(cursor, direction) {
    for (var i = 0; i < args.length; i++) {
      final result = args[i].remove(cursor, direction);
      if (result.didRemove) {
        return result;
      } else if (result.removeSelf) {
        // If this is the only argument, we will instead dismantle this command.
        // Also if this is a division and this is the second argument.
        if (args.length == 1 || (data.name == _divideCommand && i == 1)) {
          return new FragmentRemoveResult(false, dismantleCommand: true);
        } else {
          return result;
        }
      } else {
        cursor.v--;
      }
    }
    return new FragmentRemoveResult(false);
  }

  @override
  int finish(cursor, commands, ifEndsWithDigit) {
    for (final arg in args) {
      final result = arg.finish(cursor, commands, ifEndsWithDigit);
      if (result != 0) {
        return result;
      }
      cursor.v--;
    }
    return 0;
  }

  @override
  Fragment getFragmentAtCursor(cursor, direction) {
    for (final arg in args) {
      final result = arg.getFragmentAtCursor(cursor, direction);
      if (result != null) {
        return result;
      }
      cursor.v--;
    }
    return null;
  }

  @override
  String render(cursors) {
    final tex = data.latexTemplate.replaceAllMapped(_latexArgRegex, (match) {
      cursors.decrement();
      final idx = int.parse(match.group(3) ?? match.group(1));
      final argTex = args[idx].render(cursors);

      // If there is a parent operator specification in the template, add
      // parentheses by default.
      final g2 = match.group(2);
      final g4 = match.group(4);
      final opChar = g2 != null && g2.isNotEmpty ? g2 : g4;
      if (opChar != null && opChar.isNotEmpty) {
        return '\\left($argTex\\right)';
      }

      // If this argument is the first or last thing in the template string,
      // surround it with parentheses.
      else if (match.start == 0 || match.end == data.latexTemplate.length) {
        return '\\left($argTex\\right)';
      } else {
        return argTex;
      }
    });

    // Render as block or else spacing will be affected when the cursor is
    // inserted after this (since the cursor is wrapped in a block with the
    // previous element to not affect height).
    if (new RegExp(r'[\^_]').matchAsPrefix(tex) == null) {
      return '{$tex}';
    } else {
      return '$tex';
    }
  }

  @override
  String getParsableContent() {
    final str = data.parseTemplate.replaceAllMapped(_parsableArgRegex, (match) {
      final idx = int.parse(match.group(1));
      return args[idx].getParsableContent();
    });

    // Add space before and after command to make implicit multiplication
    // obvious.
    return ' $str ';
  }

  @override
  int get length => args.fold(1, (prev, arg) => prev + arg.length);
}

/// Fragment containing multiple child fragments.
class MultiFragment extends Fragment {
  static final fracBreakOps = ['=', '+', '-'];

  List<Fragment> fragments;

  MultiFragment([this.fragments]) {
    if (fragments == null) {
      fragments = new List<Fragment>();
    }
  }

  @override
  MultiFragment clone() =>
      new MultiFragment(fragments.map((frag) => frag.clone()).toList());

  /// Sequential sub and sup commands should be replaced by a composite command.
  bool _checkSequentialSubSup() {
    if (fragments.isEmpty) {
      return false;
    }

    var prev = fragments.first;
    for (var i = 1; i < fragments.length; i++) {
      final curr = fragments[i];
      if (prev is CommandFragment && curr is CommandFragment) {
        MultiFragment subArg, supArg;
        if (prev.data.name == '_' && curr.data.name == '^') {
          subArg = prev.args.first;
          supArg = curr.args.first;
        } else if (curr.data.name == '_' && prev.data.name == '^') {
          subArg = curr.args.first;
          supArg = prev.args.first;
        }

        if (subArg != null && supArg != null) {
          final command = new CommandFragment(_subSupCommand, [subArg, supArg]);
          fragments.removeRange(i - 1, i + 1);
          fragments.insert(i - 1, command);

          // In theory only one insert or delete happens at a time, so at most
          // one sequential sub/sup can be produced every time.
          return true;
        }
      }

      prev = curr;
    }

    return false;
  }

  @override
  FragmentInsertResult insert(cursor, fragment, interface) {
    for (var i = 0; i <= fragments.length; i++) {
      if (cursor.v == 0) {
        // Special case: forward slash for division, turn into frac command.
        if (fragment is CharFragment && fragment.char == '/') {
          // Get fraction numerator part.
          i--; // Go to previous fragment.
          final startI = i;
          for (; i >= 0; i--) {
            final fragment = fragments[i];
            if (fragment is CommandFragment &&
                fracBreakOps.contains(fragment.data.name)) {
              break;
            }
          }

          i++; // -1 => 0 or +/- text fragment => first fragment of numerator.

          final numerator = new MultiFragment();
          numerator.fragments.addAll(fragments.getRange(i, startI + 1));
          fragments.removeRange(i, startI + 1);

          final template = interface.resolveCommand(_divideCommand);
          if (template != null) {
            final command = new CommandFragment(template);
            command.args[0] = numerator;
            fragments.insert(i, command);
          }

          return new FragmentInsertResult(true, 2);
        } else {
          fragments.insert(i, fragment);
          if (_checkSequentialSubSup()) {
            if (fragment is CommandFragment) {
              if (fragment.data.name == '^') {
                return new FragmentInsertResult(true, 0);
              } else if (fragment.data.name == '_') {
                // Select what should be the 'subsup' command.
                // Note that the two sequential commands are turned into one.
                final prev = fragments[i - 1];
                if (prev is CommandFragment && prev.data.name == 'subsup') {
                  return new FragmentInsertResult(true, -prev.args[1].length);
                }
              }
            }
          }

          return new FragmentInsertResult(true, 1);
        }
      } else if (i < fragments.length) {
        cursor.v--;
        final result = fragments[i].insert(cursor, fragment, interface);
        if (result.didInsert) {
          return result;
        }
      }
    }
    return new FragmentInsertResult(false);
  }

  @override
  FragmentRemoveResult remove(cursor, direction) {
    for (var i = 0; i <= fragments.length; i++) {
      if (cursor.v == 0) {
        final idx = i + direction;
        if (idx >= 0 && idx < fragments.length) {
          final removed = fragments.removeAt(i + direction);
          _checkSequentialSubSup();
          return new FragmentRemoveResult(true, count: removed.length);
        } else {
          return new FragmentRemoveResult(false, removeSelf: true);
        }
      } else if (i < fragments.length) {
        final cursorOffsetBefore = cursor.v;
        cursor.v--;

        final fragment = fragments[i];
        final result = fragment.remove(cursor, direction);

        if (result.didRemove) {
          return result;
        } else {
          if (fragment is CommandFragment && result.dismantleCommand) {
            // Replace fragment with first argument in command.
            fragments.removeAt(i);
            fragments.insertAll(i, fragment.args.first.fragments);
            _checkSequentialSubSup();

            // Note, fragment.args.length is a dirty trick to get correct
            // behavior when dismantling a fraction.
            return new FragmentRemoveResult(true,
                count: fragment.args.length, removeSelf: true);
          } else if (result.removeSelf) {
            fragments.removeAt(i);
            _checkSequentialSubSup();
            return new FragmentRemoveResult(true,
                count: cursorOffsetBefore, removeSelf: true);
          }
        }
      }
    }
    return new FragmentRemoveResult(false);
  }

  @override
  int finish(cursor, interface, ifEndsWithDigit) {
    for (var i = 0; i < fragments.length; i++) {
      cursor.v--;
      if (cursor.v == 0) {
        if (ifEndsWithDigit) {
          // Only proceed if the character before the cursor is a digit.
          final prevFrag = fragments[i];
          if (prevFrag is CharFragment &&
              !new RegExp(r'^[0-9]$').hasMatch(prevFrag.char)) {
            return 0;
          }
        }

        // Append all text fragments before the cursor.
        final command = new List<int>();
        final startIdx = i;
        var allowNumbers = true;
        for (; i >= 0; i--) {
          final fragi = fragments[i];
          if (fragi is CharFragment) {
            if (fragi.char == r'\') {
              i--; // Ensure removal of backslash.
              break;
            } else if (new RegExp(r'^[0-9]$').hasMatch(fragi.char)) {
              if (!allowNumbers) {
                // Break at numbers.
                break;
              }
            } else {
              allowNumbers = false;
            }
            command.insert(0, fragi.char.codeUnitAt(0));
          } else {
            break;
          }
        }

        final commandStr = new String.fromCharCodes(command);
        final template = interface.resolveCommand(commandStr);
        if (template != null) {
          fragments.removeRange(i + 1, startIdx + 1);
          final fragment = new CommandFragment(template);
          fragments.insert(i + 1, fragment);
          return -(startIdx - i);
        } else {
          return 1;
        }
      } else {
        final result = fragments[i].finish(cursor, interface, ifEndsWithDigit);
        if (result != 0) {
          return result;
        }
      }
    }
    return 0;
  }

  @override
  Fragment getFragmentAtCursor(cursor, direction) {
    for (var i = 0; i <= fragments.length; i++) {
      if (cursor.v == 0) {
        final idx = i + direction;
        if (idx >= 0 && idx < fragments.length) {
          return fragments[idx];
        } else {
          return null;
        }
      } else if (i < fragments.length) {
        cursor.v--;
        final result = fragments[i].getFragmentAtCursor(cursor, direction);
        if (result != null) {
          return result;
        }
      }
    }
    return null;
  }

  @override
  String render(cursors) {
    final buffer = new StringBuffer();
    if (cursors.currentIsCursor()) {
      buffer.write(_cursorCommand);
      buffer.write(cursors.getColorStop());
    }

    if (fragments.isNotEmpty) {
      for (var i = 0; i < fragments.length; i++) {
        final fragment = fragments[i];
        final fragTex = fragment.render(cursors);

        cursors.decrement();
        if (cursors.currentIsCursor()) {
          if (fragment is CommandFragment &&
              new RegExp(r'[\^_]').matchAsPrefix(fragTex) == null) {
            buffer.write('{$fragTex$_cursorCommand}');
          } else {
            buffer.write('$fragTex$_cursorCommand');
          }
          buffer.write(cursors.getColorStop());
        } else {
          buffer.write(fragTex);
        }
      }
    } else {
      buffer.write(_placeholderCommand);
    }
    return buffer.toString();
  }

  List getData() {
    final list = new List();

    // Concat sequential char fragments.
    var prevIsChar = false;
    for (final fragment in fragments) {
      if (fragment is CharFragment) {
        if (prevIsChar) {
          // Concatenate
          list[list.length - 1] = '${list.last}${fragment.char}';
        } else {
          list.add(fragment.char);
          prevIsChar = true;
        }
      } else if (fragment is CommandFragment) {
        final commandData = new List()
          ..add(fragment.data.name)
          ..addAll(fragment.args.map((arg) => arg.getData()));
        list.add(commandData);
        prevIsChar = false;
      } else {
        throw new StateError(
            'MultiFragment contains fragments of type ${fragment.runtimeType}');
      }
    }

    return list;
  }

  void loadData(List data, EdiTeXInterface interface) {
    for (final item in data) {
      if (item is String) {
        // Unpack into multpile CharFragments.
        for (final char in item.runes) {
          fragments.add(new CharFragment(new String.fromCharCode(char)));
        }
      } else if (item is List) {
        // Construct [CommandData].
        final command = item.first;
        var commandData = interface.resolveInstantCommand(command);
        if (commandData == null) {
          commandData = interface.resolveCommand(command);
          if (commandData == null) {
            throw new ArgumentError("command '$command' cannot be resolved");
          }
        }

        // Construct arguments.
        final args = new List<MultiFragment>();
        for (var i = 1; i < item.length; i++) {
          args.add(new MultiFragment()..loadData(item[i], interface));
        }

        fragments.add(new CommandFragment(commandData, args));
      }
    }
  }

  @override
  String getParsableContent() =>
      fragments.fold('', (prev, frag) => '$prev${frag.getParsableContent()}');

  @override
  int get length => fragments.fold(1, (prev, arg) => prev + arg.length);
}
