// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of editex;

const _cursorCommand = r'\phantom{a}';
const _placeholderCommand = r'\phantom{b}';

class EdiTeXCommand {
  final String name, latexTemplate, parseTemplate;
  const EdiTeXCommand(this.name, this.latexTemplate, this.parseTemplate);
}

/// Interface between editor and environment. When the user enters a command
/// this function is called to find a command template. If null is returned, the
/// command is considered unknown.
typedef EdiTeXCommand CommandResolverFn(String command);

/// Same as [CommandResolverFn] but for instant commands.
typedef EdiTeXCommand InstantCommandResolverFn(String command);

/// Command used when the user types '/'. This cannot be configured. It is
/// considered a good practice to stick to LaTeX like commands when those are
/// present.
const _divideCommand = 'frac';

const _defaultCommands = const {
  'a': const EdiTeXCommand('a', r'{}_\text{?}a', 'a'),
  'a1': const EdiTeXCommand('a1', r'a_1', 'a1'),
  'a2': const EdiTeXCommand('a2', r'a_2', 'a2'),
  'frac': const EdiTeXCommand('frac', r'\frac{\,$0\,}{\,$1\,}', r'frac($0,$1)'),
  'sqrt': const EdiTeXCommand('sqrt', r'\sqrt{$0\,}', r'sqrt($0)'),
  'root': const EdiTeXCommand('root', r'\sqrt[$0]{$1\,}', r'root($0,$1)'),
  'log': const EdiTeXCommand('log', r'\log_{$0}$(1*)', r'log($0,$1)'),
  'sin': const EdiTeXCommand('sin', r'\sin$(0*)', r'sin($0)'),
  'cos': const EdiTeXCommand('cos', r'\cos$(0*)', r'cos($0)'),
  'lim': const EdiTeXCommand('lim', r'\lim_{$0\to$1}$(2*)', r'lim($0,$1,$2)'),
  'celcius': const EdiTeXCommand('celcius', r'$(0*)^\circ', r'celcius($0)'),
  'diff': const EdiTeXCommand(
      'diff', r'\frac{\partial}{\partial$(0*)}$(1*)', r'diff($0,$1)'),
  'int': const EdiTeXCommand(
      'int', r'\displaystyle\int_{$0}^{$1}$2~\text{d}$3', r'int($3,$0,$1,$2)')
};

const _defaultInstantCommands = const {
  '(': const EdiTeXCommand('(', r'\left($0\right)', r'($0)'),
  '[': const EdiTeXCommand('(', r'\left[$0\right]', r'brackets($0)'),
  '^': const EdiTeXCommand('^', r'^{$0}', r'^($0)'),
  '_': const EdiTeXCommand('_', r'_{$0}', r'_($0)'),
  '*': const EdiTeXCommand('*', r'{}\cdot{}', '*'),
  '=': const EdiTeXCommand('=', r'{}={}', '='),
  '+': const EdiTeXCommand('+', r'{}+{}', '+'),
  '-': const EdiTeXCommand('-', r'{}-{}', '-'),
  '~': const EdiTeXCommand('~', r'-', '-'),
  '!': const EdiTeXCommand('!', r'!', '!')
};

const _subSupCommand =
    const EdiTeXCommand('subsup', r'_{$0}^{$1}', r'_($0)^($1)');

EdiTeXCommand _defaultCommandResolver(String command) {
  return _defaultCommands[command];
}

EdiTeXCommand _defaultInstantCommandResolver(String command) {
  return _defaultInstantCommands[command];
}

/// Interface functions.
abstract class EdiTeXInterface {
  EdiTeXCommand resolveCommand(String command);
  EdiTeXCommand resolveInstantCommand(String command);
  bool hasCommand(String command);
  bool hasInstantCommand(String command);
}

/// Static function implementation of [EdiTeXInterface].
class StatecEdiTeXInterface implements EdiTeXInterface {
  final CommandResolverFn _resolveCommand;
  final InstantCommandResolverFn _resolveInstantCommand;

  const StatecEdiTeXInterface(
      this._resolveCommand, this._resolveInstantCommand);

  @override
  EdiTeXCommand resolveCommand(command) => _resolveCommand(command);

  @override
  EdiTeXCommand resolveInstantCommand(command) =>
      _resolveInstantCommand(command);

  @override
  bool hasCommand(String command) {
    final template = resolveCommand(command);
    return template != null;
  }

  @override
  bool hasInstantCommand(String command) {
    final template = resolveInstantCommand(command);
    return template != null;
  }
}
